@@ -1,4 +1,6 @@
-"""Virtual Table.
+"""Virtual Table handler.
+
+Hack to get local=t for a table, but without processing any events.
 """
 
 from londiste.handler import BaseHandler
@@ -6,10 +8,7 @@ from londiste.handler import BaseHandler
 __all__ = ['VirtualTableHandler', 'FakeLocalHandler']
 
 class VirtualTableHandler(BaseHandler):
-    """Virtual Table.
-
-    Hack to get local=t for a table, but without processing any events.
-    """
+    __doc__ = __doc__
     handler_name = 'vtable'
 
     def add(self, trigger_arg_list):
@@ -23,4 +22,3 @@ class FakeLocalHandler(VirtualTableHandler):
     handler_name = 'fake_local'
 
 __londiste_handlers__ = [VirtualTableHandler, FakeLocalHandler]
-
