@@ -7,6 +7,7 @@ import sys, os, re, skytools
 
 from pgq.cascade.admin import CascadeAdmin
 from londiste.exec_attrs import ExecAttrs
+from londiste.util import find_copy_source
 
 import londiste.handler
 
@@ -17,10 +18,13 @@ class LondisteSetup(CascadeAdmin):
     initial_db_name = 'node_db'
     provider_location = None
 
+    commands_without_pidfile = CascadeAdmin.commands_without_pidfile + [
+            'tables', 'seqs', 'missing', 'show-handlers']
+
     def install_code(self, db):
         self.extra_objs = [
             skytools.DBSchema("londiste", sql_file = 'londiste.sql'),
-            skytools.DBFunction("londiste.global_add_table", 2, sql_file = 'londiste.upgrade.2to3.sql'),
+            skytools.DBFunction("londiste.global_add_table", 2, sql_file = 'londiste.upgrade_2.1_to_3.1.sql'),
         ]
         CascadeAdmin.install_code(self, db)
 
@@ -36,51 +40,52 @@ class LondisteSetup(CascadeAdmin):
 
         self.set_name = self.queue_name
 
+        self.lock_timeout = self.cf.getfloat('lock_timeout', 10)
+
         londiste.handler.load_handler_modules(self.cf)
 
     def init_optparse(self, parser=None):
-        """Add londiste switches to cascadeadmin ones."""
+        """Add londiste switches to CascadeAdmin ones."""
 
         p = CascadeAdmin.init_optparse(self, parser)
         p.add_option("--expect-sync", action="store_true", dest="expect_sync",
-                    help = "no copy needed", default=False)
+                help = "no copy needed", default=False)
         p.add_option("--skip-truncate", action="store_true", dest="skip_truncate",
-                    help = "dont delete old data", default=False)
+                help = "do not delete old data", default=False)
         p.add_option("--find-copy-node", action="store_true", dest="find_copy_node",
                 help = "add: find table source for copy by walking upwards")
-        p.add_option("--copy-node", dest="copy_node",
+        p.add_option("--copy-node", metavar = "NODE", dest="copy_node",
                 help = "add: use NODE as source for initial copy")
-        p.add_option("--copy-condition", dest="copy_condition",
-                help = "copy: where expression")
         p.add_option("--force", action="store_true",
-                    help="force", default=False)
+                help="force", default=False)
         p.add_option("--all", action="store_true",
-                    help="include all tables", default=False)
+                help="include all tables", default=False)
         p.add_option("--wait-sync", action="store_true",
                 help = "add: wait until all tables are in sync"),
         p.add_option("--create", action="store_true",
-                    help="create, minimal", default=False)
+                help="create, minimal", default=False)
         p.add_option("--create-full", action="store_true",
-                    help="create, full")
+                help="create, full")
         p.add_option("--trigger-flags",
-                    help="Set trigger flags (BAIUDLQ)")
+                help="set trigger flags (BAIUDLQ)")
         p.add_option("--trigger-arg", action="append",
-                    help="Custom trigger arg")
+                help="custom trigger arg")
         p.add_option("--no-triggers", action="store_true",
-                    help="Custom trigger arg")
+                help="no triggers on table")
         p.add_option("--handler", action="store",
-                help="add: Custom handler for table")
+                help="add: custom handler for table")
         p.add_option("--handler-arg", action="append",
-                    help="add: Argument to custom handler")
+                help="add: argument to custom handler")
         p.add_option("--merge-all", action="store_true",
-                    help="merge tables from all source queues", default=False)
+                help="merge tables from all source queues", default=False)
         p.add_option("--no-merge", action="store_true",
-                    help="don't merge tables from source queues", default=False)
-        p.add_option("--max-parallel-copy", type = "int",
-                    help="max number of parallel copy processes")
-        p.add_option("--dest-table",
-                    help="add: name for actual table")
-
+                help="do not merge tables from source queues", default=False)
+        p.add_option("--max-parallel-copy", metavar = "NUM", type = "int",
+                help="max number of parallel copy processes")
+        p.add_option("--dest-table", metavar = "NAME",
+                help="add: name for actual table")
+        p.add_option("--skip-non-existing", action="store_true",
+                help="add: skip object that does not exist")
         return p
 
     def extra_init(self, node_type, node_db, provider_db):
@@ -114,6 +119,13 @@ class LondisteSetup(CascadeAdmin):
     def is_root(self):
         return self.queue_info.local_node.type == 'root'
 
+    def set_lock_timeout(self, curs):
+        ms = int(1000 * self.lock_timeout)
+        if ms > 0:
+            q = "SET LOCAL statement_timeout = %d" % ms
+            self.log.debug(q)
+            curs.execute(q)
+
     def cmd_add_table(self, *args):
         """Attach table(s) to local node."""
 
@@ -137,27 +149,6 @@ class LondisteSetup(CascadeAdmin):
         needs_tbl = self.handler_needs_table()
         args = self.expand_arg_list(dst_db, 'r', False, args, needs_tbl)
 
-        # search for usable copy node if requested
-        if (self.options.find_copy_node
-                and not self.is_root()
-                and needs_tbl):
-            src_db = self.find_copy_node(dst_db, args)
-            src_curs = src_db.cursor()
-            src_tbls = self.fetch_set_tables(src_curs)
-            src_db.commit()
-
-        # dont check for exist/not here (root handling)
-        if not self.is_root() and not self.options.expect_sync:
-            problems = False
-            for tbl in args:
-                tbl = skytools.fq_name(tbl)
-                if (tbl in src_tbls) and not src_tbls[tbl]['local']:
-                    self.log.error("Table %s does not exist on provider, need to switch to different provider" % tbl)
-                    problems = True
-            if problems:
-                self.log.error("Problems, canceling operation")
-                sys.exit(1)
-
         # pick proper create flags
         if self.options.create_full:
             create_flags = skytools.T_ALL
@@ -166,6 +157,32 @@ class LondisteSetup(CascadeAdmin):
         else:
             create_flags = 0
 
+        # search for usable copy node if requested & needed
+        if (self.options.find_copy_node and create_flags != 0
+                and needs_tbl and not self.is_root()):
+            src_name, src_loc, _ = find_copy_source(self, self.queue_name, args, None, self.provider_location)
+            self.options.copy_node = src_name
+            self.close_database('provider_db')
+            src_db = self.get_provider_db()
+            src_curs = src_db.cursor()
+            src_tbls = self.fetch_set_tables(src_curs)
+            src_db.commit()
+
+        # dont check for exist/not here (root handling)
+        if not self.is_root() and not self.options.expect_sync and not self.options.find_copy_node:
+            problems = False
+            for tbl in args:
+                tbl = skytools.fq_name(tbl)
+                if (tbl in src_tbls) and not src_tbls[tbl]['local']:
+                    if self.options.skip_non_existing:
+                        self.log.warning("Table %s does not exist on provider", tbl)
+                    else:
+                        self.log.error("Table %s does not exist on provider, need to switch to different provider", tbl)
+                        problems = True
+            if problems:
+                self.log.error("Problems, canceling operation")
+                sys.exit(1)
+
         # sanity check
         if self.options.dest_table and len(args) > 1:
             self.log.error("--dest-table can be given only for single table")
@@ -179,7 +196,6 @@ class LondisteSetup(CascadeAdmin):
         if self.options.wait_sync:
             self.wait_for_sync(dst_db)
 
-
     def add_table(self, src_db, dst_db, tbl, create_flags, src_tbls):
         # use full names
         tbl = skytools.fq_name(tbl)
@@ -189,6 +205,9 @@ class LondisteSetup(CascadeAdmin):
         src_curs = src_db.cursor()
         dst_curs = dst_db.cursor()
         tbl_exists = skytools.exists_table(dst_curs, dest_table)
+        dst_db.commit()
+
+        self.set_lock_timeout(dst_curs)
 
         if dest_table == tbl:
             desc = tbl
@@ -197,12 +216,12 @@ class LondisteSetup(CascadeAdmin):
 
         if create_flags:
             if tbl_exists:
-                self.log.info('Table %s already exist, not touching' % desc)
+                self.log.info('Table %s already exist, not touching', desc)
             else:
                 src_dest_table = src_tbls[tbl]['dest_table']
                 if not skytools.exists_table(src_curs, src_dest_table):
                     # table not present on provider - nowhere to get the DDL from
-                    self.log.warning('Table %s missing on provider, cannot create, skipping' % desc)
+                    self.log.warning('Table %s missing on provider, cannot create, skipping', desc)
                     return
                 schema = skytools.fq_name_parts(dest_table)[0]
                 if not skytools.exists_schema(dst_curs, schema):
@@ -216,39 +235,25 @@ class LondisteSetup(CascadeAdmin):
                 if src_dest_table != dest_table:
                     newname = dest_table
                 s.create(dst_curs, create_flags, log = self.log, new_table_name = newname)
+        elif not tbl_exists and self.options.skip_non_existing:
+            self.log.warning('Table %s does not exist on local node, skipping', desc)
+            return
 
-        tgargs = []
-        if self.options.trigger_arg:
-            tgargs = self.options.trigger_arg
-        tgflags = self.options.trigger_flags
-        if tgflags:
-            tgargs.append('tgflags='+tgflags)
-        if self.options.no_triggers:
-            tgargs.append('no_triggers')
-        if self.options.merge_all:
-            tgargs.append('merge_all')
-        if self.options.no_merge:
-            tgargs.append('no_merge')
+        tgargs = self.build_tgargs()
 
         attrs = {}
+
         if self.options.handler:
-            hstr = londiste.handler.create_handler_string(
-                            self.options.handler, self.options.handler_arg)
-            p = londiste.handler.build_handler(tbl, hstr, self.options.dest_table)
-            attrs['handler'] = hstr
-            p.add(tgargs)
+            attrs['handler'] = self.build_handler(tbl, tgargs, self.options.dest_table)
 
-        if self.options.copy_node:
+        if self.options.find_copy_node:
+            attrs['copy_node'] = '?'
+        elif self.options.copy_node:
             attrs['copy_node'] = self.options.copy_node
 
-        if self.options.expect_sync:
-            tgargs.append('expect_sync')
-
         if not self.options.expect_sync:
             if self.options.skip_truncate:
                 attrs['skip_truncate'] = 1
-            if self.options.copy_condition:
-                attrs['copy_condition'] = self.options.copy_condition
 
         if self.options.max_parallel_copy:
             attrs['max_parallel_copy'] = self.options.max_parallel_copy
@@ -263,6 +268,32 @@ class LondisteSetup(CascadeAdmin):
         self.exec_cmd(dst_curs, q, args)
         dst_db.commit()
 
+    def build_tgargs(self):
+        """Build trigger args"""
+        tgargs = []
+        if self.options.trigger_arg:
+            tgargs = self.options.trigger_arg
+        tgflags = self.options.trigger_flags
+        if tgflags:
+            tgargs.append('tgflags='+tgflags)
+        if self.options.no_triggers:
+            tgargs.append('no_triggers')
+        if self.options.merge_all:
+            tgargs.append('merge_all')
+        if self.options.no_merge:
+            tgargs.append('no_merge')
+        if self.options.expect_sync:
+            tgargs.append('expect_sync')
+        return tgargs
+
+    def build_handler(self, tbl, tgargs, dest_table=None):
+        """Build handler and return handler string"""
+        hstr = londiste.handler.create_handler_string(
+                self.options.handler, self.options.handler_arg)
+        p = londiste.handler.build_handler(tbl, hstr, dest_table)
+        p.add(tgargs)
+        return hstr
+
     def handler_needs_table(self):
         if self.options.handler:
             hstr = londiste.handler.create_handler_string(
@@ -271,20 +302,11 @@ class LondisteSetup(CascadeAdmin):
             return p.needs_table()
         return True
 
-    def handler_allows_copy(self, table_attrs):
-        """Decide if table is copyable based on attrs."""
-        if not table_attrs:
-            return True
-        attrs = skytools.db_urldecode(table_attrs)
-        hstr = attrs['handler']
-        p = londiste.handler.build_handler('unused.string', hstr, None)
-        return p.needs_table()
-
     def sync_table_list(self, dst_curs, src_tbls, dst_tbls):
         for tbl in src_tbls.keys():
             q = "select * from londiste.global_add_table(%s, %s)"
             if tbl not in dst_tbls:
-                self.log.info("Table %s info missing from subscriber, adding" % tbl)
+                self.log.info("Table %s info missing from subscriber, adding", tbl)
                 self.exec_cmd(dst_curs, q, [self.set_name, tbl])
                 dst_tbls[tbl] = {'local': False, 'dest_table': tbl}
         for tbl in dst_tbls.keys():
@@ -311,6 +333,50 @@ class LondisteSetup(CascadeAdmin):
         q = "select * from londiste.local_remove_table(%s, %s)"
         self.exec_cmd_many(db, q, [self.set_name], args)
 
+    def cmd_change_handler(self, tbl):
+        """Change handler (table_attrs) of the replicated table."""
+
+        self.load_local_info()
+
+        tbl = skytools.fq_name(tbl)
+
+        db = self.get_database('db')
+        curs = db.cursor()
+        q = "select table_attrs, dest_table "\
+            " from londiste.get_table_list(%s) "\
+            " where table_name = %s and local"
+        curs.execute(q, [self.set_name, tbl])
+        if curs.rowcount == 0:
+            self.log.error("Table %s not found on this node", tbl)
+            sys.exit(1)
+
+        attrs, dest_table = curs.fetchone()
+        attrs = skytools.db_urldecode(attrs or '')
+        old_handler = attrs.get('handler')
+
+        tgargs = self.build_tgargs()
+        if self.options.handler:
+            new_handler = self.build_handler(tbl, tgargs, dest_table)
+        else:
+            new_handler = None
+
+        if old_handler == new_handler:
+            self.log.info("Handler is already set to desired value, nothing done")
+            sys.exit(0)
+
+        if new_handler:
+            attrs['handler'] = new_handler
+        elif 'handler' in attrs:
+            del attrs['handler']
+
+        args = [self.set_name, tbl, tgargs, None]
+        if attrs:
+            args[3] = skytools.db_urlencode(attrs)
+
+        q = "select * from londiste.local_change_handler(%s, %s, %s, %s)"
+        self.exec_cmd(curs, q, args)
+        db.commit()
+
     def cmd_add_seq(self, *args):
         """Attach seqs(s) to local node."""
         dst_db = self.get_database('db')
@@ -346,18 +412,21 @@ class LondisteSetup(CascadeAdmin):
         seq_exists = skytools.exists_sequence(dst_curs, seq)
         if create_flags:
             if seq_exists:
-                self.log.info('Sequence %s already exist, not creating' % seq)
+                self.log.info('Sequence %s already exist, not creating', seq)
             else:
                 if not skytools.exists_sequence(src_curs, seq):
                     # sequence not present on provider - nowhere to get the DDL from
-                    self.log.warning('Sequence "%s" missing on provider, skipping' % seq)
+                    self.log.warning('Sequence "%s" missing on provider, skipping', seq)
                     return
                 s = skytools.SeqStruct(src_curs, seq)
                 src_db.commit()
                 s.create(dst_curs, create_flags, log = self.log)
         elif not seq_exists:
-            self.log.warning('Sequence "%s" missing on subscriber, use --create if necessary' % seq)
-            return
+            if self.options.skip_non_existing:
+                self.log.warning('Sequence "%s" missing on local node, skipping', seq)
+                return
+            else:
+                raise skytools.UsageError("Sequence %r missing on local node", seq)
 
         q = "select * from londiste.local_add_seq(%s, %s)"
         self.exec_cmd(dst_curs, q, [self.set_name, seq])
@@ -374,7 +443,7 @@ class LondisteSetup(CascadeAdmin):
         for seq in src_seqs.keys():
             q = "select * from londiste.global_update_seq(%s, %s, %s)"
             if seq not in dst_seqs:
-                self.log.info("Sequence %s info missing from subscriber, adding" % seq)
+                self.log.info("Sequence %s info missing from subscriber, adding", seq)
                 self.exec_cmd(dst_curs, q, [self.set_name, seq, src_seqs[seq]['last_value']])
                 tmp = src_seqs[seq].copy()
                 tmp['local'] = False
@@ -394,9 +463,45 @@ class LondisteSetup(CascadeAdmin):
         self.exec_cmd_many(db, q, [self.set_name], args)
 
     def cmd_resync(self, *args):
-        """Reload data from provider node.."""
+        """Reload data from provider node."""
         db = self.get_database('db')
         args = self.expand_arg_list(db, 'r', True, args)
+
+        if not self.options.find_copy_node:
+            self.load_local_info()
+            src_db = self.get_provider_db()
+            src_curs = src_db.cursor()
+            src_tbls = self.fetch_set_tables(src_curs)
+            src_db.commit()
+
+            problems = 0
+            for tbl in args:
+                tbl = skytools.fq_name(tbl)
+                if tbl not in src_tbls or not src_tbls[tbl]['local']:
+                    self.log.error("Table %s does not exist on provider, need to switch to different provider", tbl)
+                    problems += 1
+            if problems > 0:
+                self.log.error("Problems, cancelling operation")
+                sys.exit(1)
+
+        if self.options.find_copy_node or self.options.copy_node:
+            q = "select table_name, table_attrs from londiste.get_table_list(%s) where local"
+            cur = db.cursor()
+            cur.execute(q, [self.set_name])
+            for row in cur.fetchall():
+                if row['table_name'] not in args:
+                    continue
+                attrs = skytools.db_urldecode (row['table_attrs'] or '')
+
+                if self.options.find_copy_node:
+                    attrs['copy_node'] = '?'
+                elif self.options.copy_node:
+                    attrs['copy_node'] = self.options.copy_node
+
+                attrs = skytools.db_urlencode (attrs)
+                q = "select * from londiste.local_set_table_attrs (%s, %s, %s)"
+                self.exec_cmd(db, q, [self.set_name, row['table_name'], attrs])
+
         q = "select * from londiste.local_set_table_state(%s, %s, null, null)"
         self.exec_cmd_many(db, q, [self.set_name], args)
 
@@ -467,8 +572,8 @@ class LondisteSetup(CascadeAdmin):
             q = "select * from londiste.execute_start(%s, %s, %s, true, %s)"
             res = self.exec_cmd(db, q, [self.queue_name, fname, sql, attrs.to_urlenc()], commit = False)
             ret = res[0]['ret_code']
-            if ret >= 300:
-                self.log.warning("Skipping execution of '%s'" % fname)
+            if ret > 200:
+                self.log.warning("Skipping execution of '%s'", fname)
                 continue
             if attrs.need_execute(curs, local_tables, local_seqs):
                 self.log.info("%s: executing sql", fname)
@@ -481,64 +586,9 @@ class LondisteSetup(CascadeAdmin):
             self.exec_cmd(db, q, [self.queue_name, fname], commit = False)
         db.commit()
 
-    def find_copy_node(self, dst_db, args):
-        src_db = self.get_provider_db()
-
-        need = {}
-        for t in args:
-            need[t] = 1
-
-        while 1:
-            src_curs = src_db.cursor()
-
-            q = "select * from pgq_node.get_node_info(%s)"
-            src_curs.execute(q, [self.queue_name])
-            info = src_curs.fetchone()
-            if info['ret_code'] >= 400:
-                raise UsageError("Node does not exists")
-
-            self.log.info("Checking if %s can be used for copy", info['node_name'])
-
-            q = "select table_name, local, table_attrs from londiste.get_table_list(%s)"
-            src_curs.execute(q, [self.queue_name])
-            got = {}
-            for row in src_curs.fetchall():
-                tbl = row['table_name']
-                if tbl not in need:
-                    continue
-                if not row['local']:
-                    self.log.debug("Problem: %s is not local", tbl)
-                    continue
-                if not self.handler_allows_copy(row['table_attrs']):
-                    self.log.debug("Problem: %s handler does not store data [%s]", tbl, row['table_attrs'])
-                    continue
-                self.log.debug("Good: %s is usable", tbl)
-                got[row['table_name']] = 1
-
-            ok = 1
-            for t in args:
-                if t not in got:
-                    self.log.info("Node %s does not have all tables", info['node_name'])
-                    ok = 0
-                    break
-
-            if ok:
-                self.options.copy_node = info['node_name']
-                self.log.info("Node %s seems good source, using it", info['node_name'])
-                break
-
-            if info['node_type'] == 'root':
-                raise skytools.UsageError("Found root and no source found")
-
-            self.close_database('provider_db')
-            src_db = self.get_database('provider_db', connstr = info['provider_location'])
-
-        return src_db
-
     def get_provider_db(self):
-
-        # use custom node for copy
         if self.options.copy_node:
+            # use custom node for copy
             source_node = self.options.copy_node
             m = self.queue_info.get_member(source_node)
             if not m:
@@ -552,8 +602,8 @@ class LondisteSetup(CascadeAdmin):
             q = 'select * from pgq_node.get_node_info(%s)'
             res = self.exec_cmd(db, q, [self.queue_name], quiet = True)
             self.provider_location = res[0]['provider_location']
-        return self.get_database('provider_db', connstr = self.provider_location)
 
+        return self.get_database('provider_db', connstr = self.provider_location, profile = 'remote')
 
     def expand_arg_list(self, db, kind, existing, args, needs_tbl=True):
         curs = db.cursor()
@@ -588,14 +638,15 @@ class LondisteSetup(CascadeAdmin):
             else:
                 return lst_missing
 
-
         allow_nonexist = not needs_tbl
         if existing:
             res = self.solve_globbing(args, lst_exists, map_exists, map_missing, allow_nonexist)
         else:
             res = self.solve_globbing(args, lst_missing, map_missing, map_exists, allow_nonexist)
-        return res
 
+        if not res:
+            self.log.info("what to do ?")
+        return res
 
     def solve_globbing(self, args, full_list, full_map, reverse_map, allow_nonexist):
         def glob2regex(s):
@@ -623,16 +674,16 @@ class LondisteSetup(CascadeAdmin):
                     res_list.append(a)
                     res_map[a] = 1
                 elif a in reverse_map:
-                    self.log.info("%s already processed" % a)
+                    self.log.info("%s already processed", a)
                 elif allow_nonexist:
                     res_list.append(a)
                     res_map[a] = 1
                 elif self.options.force:
-                    self.log.warning("%s not available, but --force is used" % a)
+                    self.log.warning("%s not available, but --force is used", a)
                     res_list.append(a)
                     res_map[a] = 1
                 else:
-                    self.log.warning("%s not available" % a)
+                    self.log.warning("%s not available", a)
                     err = 1
         if err:
             raise skytools.UsageError("Cannot proceed")
@@ -640,6 +691,7 @@ class LondisteSetup(CascadeAdmin):
 
     def load_extra_status(self, curs, node):
         """Fetch extra info."""
+        # must be thread-safe (!)
         CascadeAdmin.load_extra_status(self, curs, node)
         curs.execute("select * from londiste.get_table_list(%s)", [self.queue_name])
         n_ok = n_half = n_ign = 0
@@ -665,18 +717,19 @@ class LondisteSetup(CascadeAdmin):
         dst_curs = dst_db.cursor()
 
         partial = {}
-        done_pos = 1
         startup_info = 0
         while 1:
             dst_curs.execute(q, [self.queue_name])
             rows = dst_curs.fetchall()
             dst_db.commit()
 
+            total_count = 0
             cur_count = 0
             done_list = []
             for row in rows:
                 if not row['local']:
                     continue
+                total_count += 1
                 tbl = row['table_name']
                 if row['merge_state'] != 'ok':
                     partial[tbl] = 0
@@ -686,13 +739,14 @@ class LondisteSetup(CascadeAdmin):
                         partial[tbl] = 1
                         done_list.append(tbl)
 
+            done_count = total_count - cur_count
+
             if not startup_info:
-                self.log.info("%d table(s) to copy", len(partial))
+                self.log.info("%d/%d table(s) to copy", cur_count, total_count)
                 startup_info = 1
 
             for done in done_list:
-                self.log.info("%s: finished (%d/%d)", done, done_pos, len(partial))
-                done_pos += 1
+                self.log.info("%s: finished (%d/%d)", done, done_count, total_count)
 
             if cur_count == 0:
                 break
@@ -701,3 +755,36 @@ class LondisteSetup(CascadeAdmin):
 
         self.log.info("All done")
 
+    def resurrect_dump_event(self, ev, stats, batch_info):
+        """Collect per-table stats."""
+
+        super(LondisteSetup, self).resurrect_dump_event(ev, stats, batch_info)
+
+        ROLLBACK = 'can rollback'
+        NO_ROLLBACK = 'cannot rollback'
+
+        if ev.ev_type == 'TRUNCATE':
+            if 'truncated_tables' not in stats:
+                stats['truncated_tables'] = []
+            tlist = stats['truncated_tables']
+            tbl = ev.ev_extra1
+            if tbl not in tlist:
+                tlist.append(tbl)
+        elif ev.ev_type[:2] in ('I:', 'U:', 'D:', 'I', 'U', 'D'):
+            op = ev.ev_type[0]
+            tbl = ev.ev_extra1
+            bak = ev.ev_extra3
+            tblkey = 'table: %s' % tbl
+            if tblkey not in stats:
+                stats[tblkey] = [0,0,0,ROLLBACK]
+            tinfo = stats[tblkey]
+            if op == 'I':
+                tinfo[0] += 1
+            elif op == 'U':
+                tinfo[1] += 1
+                if not bak:
+                    tinfo[3] = NO_ROLLBACK
+            elif op == 'D':
+                tinfo[2] += 1
+                if not bak and ev.ev_type == 'D':
+                    tinfo[3] = NO_ROLLBACK
