@@ -7,6 +7,7 @@ For internal usage.
 
 import sys, time, skytools
 
+from londiste.util import find_copy_source
 from skytools.dbstruct import *
 from londiste.playback import *
 
@@ -63,7 +64,7 @@ class CopyTable(Replicator):
             if tbl_stat.copy_role == 'wait-copy':
                 self.log.info('waiting for first partition to initialize copy')
             elif tbl_stat.max_parallel_copies_reached():
-                self.log.info('number of max parallel copies (%s) reached' %\
+                self.log.info('number of max parallel copies (%s) reached',
                                 tbl_stat.max_parallel_copy)
             else:
                 break
@@ -80,7 +81,7 @@ class CopyTable(Replicator):
             if pt.state == TABLE_OK:
                 break
 
-            self.log.warning("table %s not in sync yet on provider, waiting" % tbl_stat.name)
+            self.log.warning("table %s not in sync yet on provider, waiting", tbl_stat.name)
             time.sleep(10)
 
         src_real_table = pt.dest_table
@@ -101,7 +102,7 @@ class CopyTable(Replicator):
 
         self.sync_database_encodings(src_db, dst_db)
 
-        self.log.info("Starting full copy of %s" % tbl_stat.name)
+        self.log.info("Starting full copy of %s", tbl_stat.name)
 
         # just in case, drop all fkeys (in case "replay" was skipped)
         # !! this may commit, so must be done before anything else !!
@@ -123,14 +124,14 @@ class CopyTable(Replicator):
         common_cols = []
         for c in slist:
             if c not in dlist:
-                self.log.warning("Table %s column %s does not exist on subscriber"
-                                 % (tbl_stat.name, c))
+                self.log.warning("Table %s column %s does not exist on subscriber",
+                                 tbl_stat.name, c)
             else:
                 common_cols.append(c)
         for c in dlist:
             if c not in slist:
-                self.log.warning("Table %s column %s does not exist on provider"
-                                 % (tbl_stat.name, c))
+                self.log.warning("Table %s column %s does not exist on provider",
+                                 tbl_stat.name, c)
 
         # drop unnecessary stuff
         if cmode > 0:
@@ -139,9 +140,9 @@ class CopyTable(Replicator):
 
             # drop data
             if tbl_stat.table_attrs.get('skip_truncate'):
-                self.log.info("%s: skipping truncate" % tbl_stat.name)
+                self.log.info("%s: skipping truncate", tbl_stat.name)
             else:
-                self.log.info("%s: truncating" % tbl_stat.name)
+                self.log.info("%s: truncating", tbl_stat.name)
                 q = "truncate "
                 if dst_db.server_version >= 80400:
                     q += "only "
@@ -159,7 +160,12 @@ class CopyTable(Replicator):
                 tbl_stat.dropped_ddl = ddl
 
         # do truncate & copy
-        self.real_copy(src_curs, dst_curs, tbl_stat, common_cols, src_real_table)
+        self.log.info("%s: start copy", tbl_stat.name)
+        p = tbl_stat.get_plugin()
+        stats = p.real_copy(src_real_table, src_curs, dst_curs, common_cols)
+        if stats:
+            self.log.info("%s: copy finished: %d bytes, %d rows",
+                          tbl_stat.name, stats[0], stats[1])
 
         # get snapshot
         src_curs.execute("select txid_current_snapshot()")
@@ -208,29 +214,13 @@ class CopyTable(Replicator):
         if tbl_stat.copy_role == 'wait-replay':
             return
 
-        # if copy done, request immidiate tick from pgqadm,
+        # if copy done, request immediate tick from pgqd,
         # to make state juggling faster.  on mostly idle db-s
         # each step may take tickers idle_timeout secs, which is pain.
         q = "select pgq.force_tick(%s)"
         src_curs.execute(q, [self.queue_name])
         src_db.commit()
 
-    def real_copy(self, srccurs, dstcurs, tbl_stat, col_list, src_real_table):
-        "Actual copy."
-
-        tablename = tbl_stat.name
-        # do copy
-        self.log.info("%s: start copy" % tablename)
-        p = tbl_stat.get_plugin()
-        cond_list = []
-        cond = tbl_stat.table_attrs.get('copy_condition')
-        if cond:
-            cond_list.append(cond)
-        stats = p.real_copy(src_real_table, srccurs, dstcurs, col_list, cond_list)
-        if stats:
-            self.log.info("%s: copy finished: %d bytes, %d rows" % (
-                          tablename, stats[0], stats[1]))
-
     def work(self):
         if not self.reg_ok:
             # check if needed? (table, not existing reg)
@@ -252,22 +242,26 @@ class CopyTable(Replicator):
             if v_attrs:
                 attrs = skytools.db_urldecode(v_attrs)
 
+        # fetch parent consumer state
+        q = "select * from pgq_node.get_consumer_state(%s, %s)"
+        rows = self.exec_cmd(dst_db, q, [ self.queue_name, self.old_consumer_name ])
+        state = rows[0]
+        source_node = state['provider_node']
+        source_location = state['provider_location']
+
         # do we have node here?
         if 'copy_node' in attrs:
-            # take node from attrs
-            source_node = attrs['copy_node']
-            q = "select * from pgq_node.get_queue_locations(%s) where node_name = %s"
-            dst_curs.execute(q, [ self.queue_name, source_node ])
-            rows = dst_curs.fetchall()
-            if len(rows):
-                source_location = rows[0]['node_location']
-        else:
-            # fetch parent consumer state
-            q = "select * from pgq_node.get_consumer_state(%s, %s)"
-            rows = self.exec_cmd(dst_db, q, [ self.queue_name, self.old_consumer_name ])
-            state = rows[0]
-            source_node = state['provider_node']
-            source_location = state['provider_location']
+            if attrs['copy_node'] == '?':
+                source_node, source_location, wname = find_copy_source(self,
+                        self.queue_name, self.copy_table_name, source_node, source_location)
+            else:
+                # take node from attrs
+                source_node = attrs['copy_node']
+                q = "select * from pgq_node.get_queue_locations(%s) where node_name = %s"
+                dst_curs.execute(q, [ self.queue_name, source_node ])
+                rows = dst_curs.fetchall()
+                if len(rows):
+                    source_location = rows[0]['node_location']
 
         self.log.info("Using '%s' as source node", source_node)
         self.register_consumer(source_location)
@@ -275,4 +269,3 @@ class CopyTable(Replicator):
 if __name__ == '__main__':
     script = CopyTable(sys.argv[1:])
     script.start()
-
