@@ -10,8 +10,10 @@ create or replace function londiste.upgrade_schema()
 returns int4 as $$
 -- updates table structure if necessary
 declare
+    pgversion int;
     cnt int4 = 0;
 begin
+    show server_version_num into pgversion;
 
     -- table_info: check (dropped_ddl is null or merge_state in ('in-copy', 'catching-up'))
     perform 1 from information_schema.check_constraints
@@ -35,6 +37,26 @@ begin
         alter table londiste.table_info add column dest_table text;
     end if;
 
+    -- table_info: change trigger timing
+    if pgversion >= 90100 then
+        perform 1 from information_schema.triggers
+          where event_object_schema = 'londiste'
+            and event_object_table = 'table_info'
+            and trigger_name = 'table_info_trigger_sync'
+            and action_timing = 'AFTER';
+    else
+        perform 1 from information_schema.triggers
+          where event_object_schema = 'londiste'
+            and event_object_table = 'table_info'
+            and trigger_name = 'table_info_trigger_sync'
+            and condition_timing = 'AFTER';
+    end if;
+    if found then
+        drop trigger table_info_trigger_sync on londiste.table_info;
+        create trigger table_info_trigger_sync before delete on londiste.table_info
+            for each row execute procedure londiste.table_info_trigger();
+    end if;
+
     -- applied_execute.dest_table
     perform 1 from information_schema.columns
       where table_schema = 'londiste'
@@ -420,6 +442,264 @@ $$ language plpgsql;
 
 
 
+create or replace function londiste.create_trigger(
+    in i_queue_name     text,
+    in i_table_name     text,
+    in i_trg_args       text[],
+    in i_dest_table     text,
+    in i_node_type      text,
+    out ret_code        int4,
+    out ret_note        text,
+    out trigger_name    text)
+as $$
+------------------------------------------------------------------------
+-- Function: londiste.create_trigger(5)
+--
+--     Create or replace londiste trigger(s)
+--
+-- Parameters:
+--      i_queue_name - queue name
+--      i_table_name - table name
+--      i_trg_args   - args to trigger
+--      i_dest_table - actual name of destination table (NULL if same as src)
+--      i_node_type  - l3 node type
+--
+-- Trigger args:
+--      See documentation for pgq triggers.
+--
+-- Trigger creation flags (default: AIUDL):
+--      I - ON INSERT
+--      U - ON UPDATE
+--      D - ON DELETE
+--      Q - use pgq.sqltriga() as trigger function
+--      L - use pgq.logutriga() as trigger function
+--      B - BEFORE
+--      A - AFTER
+--      S - SKIP
+--
+-- Returns:
+--      200 - Ok
+--      201 - Trigger not created
+--      405 - Multiple SKIP triggers
+--
+------------------------------------------------------------------------
+declare
+    trigger_name text;
+    lg_func text;
+    lg_pos text;
+    lg_event text;
+    lg_args text[];
+    _old_tgargs bytea;
+    _new_tgargs bytea;
+    trunctrg_name text;
+    pgversion int;
+    sql text;
+    arg text;
+    i integer;
+    _extra_args text[] := '{}';
+    -- skip trigger
+    _skip_prefix text := 'zzz_';
+    _skip_trg_count integer;
+    _skip_trg_name text;
+    -- given tgflags array
+    _tgflags char[];
+    -- ordinary argument array
+    _args text[];
+    -- array with all valid tgflags values
+    _valid_flags char[] := array['B','A','Q','L','I','U','D','S'];
+    -- argument flags
+    _skip boolean := false;
+    _no_triggers boolean := false;
+    _got_extra1 boolean := false;
+begin
+    -- parse trigger args
+    if array_lower(i_trg_args, 1) is not null then
+        for i in array_lower(i_trg_args, 1) .. array_upper(i_trg_args, 1) loop
+            arg := i_trg_args[i];
+            if arg like 'tgflags=%' then
+                -- special flag handling
+                arg := upper(substr(arg, 9));
+                for j in array_lower(_valid_flags, 1) .. array_upper(_valid_flags, 1) loop
+                    if position(_valid_flags[j] in arg) > 0 then
+                        _tgflags := array_append(_tgflags, _valid_flags[j]);
+                    end if;
+                end loop;
+            elsif arg = 'no_triggers' then
+                _no_triggers := true;
+            elsif lower(arg) = 'skip' then
+                _skip := true;
+            elsif arg = 'virtual_table' then
+                _no_triggers := true;   -- do not create triggers
+            elsif arg not in ('expect_sync', 'skip_truncate', 'merge_all', 'no_merge') then -- ignore add-table args
+                if arg like 'ev_extra1=%' then
+                    _got_extra1 := true;
+                end if;
+                -- ordinary arg
+                _args = array_append(_args, quote_literal(arg));
+            end if;
+        end loop;
+    end if;
+
+    if i_dest_table <> i_table_name and not _got_extra1 then
+        -- if renamed table, enforce trigger to put
+        -- global table name into extra1
+        arg := 'ev_extra1=' || quote_literal(i_table_name);
+        _args := array_append(_args, quote_literal(arg));
+    end if;
+
+    trigger_name := '_londiste_' || i_queue_name;
+    lg_func := 'pgq.logutriga';
+    lg_event := '';
+    lg_args := array[quote_literal(i_queue_name)];
+    lg_pos := 'after';
+
+    if array_lower(_args, 1) is not null then
+        lg_args := lg_args || _args;
+    end if;
+
+    if 'B' = any(_tgflags) then
+        lg_pos := 'before';
+    end if;
+    if 'A' = any(_tgflags)  then
+        lg_pos := 'after';
+    end if;
+    if 'Q' = any(_tgflags) then
+        lg_func := 'pgq.sqltriga';
+    end if;
+    if 'L' = any(_tgflags) then
+        lg_func := 'pgq.logutriga';
+    end if;
+    if 'I' = any(_tgflags) then
+        lg_event := lg_event || ' or insert';
+    end if;
+    if 'U' = any(_tgflags) then
+        lg_event := lg_event || ' or update';
+    end if;
+    if 'D' = any(_tgflags) then
+        lg_event := lg_event || ' or delete';
+    end if;
+    if 'S' = any(_tgflags) then
+        _skip := true;
+    end if;
+
+    if i_node_type = 'leaf' then
+        -- on weird leafs the trigger funcs may not exist
+        perform 1 from pg_proc p join pg_namespace n on (n.oid = p.pronamespace)
+            where n.nspname = 'pgq' and p.proname in ('logutriga', 'sqltriga');
+        if not found then
+            select 201, 'Trigger not created' into ret_code, ret_note;
+            return;
+        end if;
+        -- on regular leaf, install deny trigger
+        _extra_args := array_append(_extra_args, quote_literal('deny'));
+    end if;
+
+    if _skip or lg_pos = 'after' then
+        -- get count and name of existing skip triggers
+        select count(*), min(t.tgname)
+            into _skip_trg_count, _skip_trg_name
+            from pg_catalog.pg_trigger t
+            where t.tgrelid = londiste.find_table_oid(i_dest_table)
+                and position(E'\\000SKIP\\000'::bytea in tgargs) > 0;
+    end if;
+
+    -- make sure new trigger won't be effectively inactive
+    if lg_pos = 'after' and _skip_trg_count > 0 then
+        select 403, 'AFTER trigger cannot work with SKIP trigger(s)'
+        into ret_code, ret_note;
+        return;
+    end if;
+
+    -- if skip param given, rename previous skip triggers and prefix current
+    if _skip then
+        -- if no previous skip triggers, prefix name and add SKIP to args
+        if _skip_trg_count = 0 then
+            trigger_name := _skip_prefix || trigger_name;
+            lg_args := array_append(lg_args, quote_literal('SKIP'));
+        -- if one previous skip trigger, check it's prefix and
+        -- do not use SKIP on current trigger
+        elsif _skip_trg_count = 1 then
+            -- if not prefixed then rename
+            if position(_skip_prefix in _skip_trg_name) != 1 then
+                sql := 'alter trigger ' || _skip_trg_name
+                    || ' on ' || londiste.quote_fqname(i_dest_table)
+                    || ' rename to ' || _skip_prefix || _skip_trg_name;
+                execute sql;
+            end if;
+        else
+            select 405, 'Multiple SKIP triggers'
+            into ret_code, ret_note;
+            return;
+        end if;
+    end if;
+
+    -- create Ins/Upd/Del trigger if it does not exists already
+    select t.tgargs
+        from pg_catalog.pg_trigger t
+        where t.tgrelid = londiste.find_table_oid(i_dest_table)
+            and t.tgname = trigger_name
+        into _old_tgargs;
+
+    if found then
+        _new_tgargs := decode(lg_args[1], 'escape');
+        for i in 2 .. array_upper(lg_args, 1) loop
+            _new_tgargs := _new_tgargs || E'\\000'::bytea || decode(lg_args[i], 'escape');
+        end loop;
+
+        if _old_tgargs is distinct from _new_tgargs then
+            sql := 'drop trigger if exists ' || quote_ident(trigger_name)
+                || ' on ' || londiste.quote_fqname(i_dest_table);
+            execute sql;
+        end if;
+    end if;
+
+    if not found or _old_tgargs is distinct from _new_tgargs then
+        if _no_triggers then
+            select 201, 'Trigger not created'
+            into ret_code, ret_note;
+            return;
+        end if;
+
+        -- finalize event
+        lg_event := substr(lg_event, 4); -- remove ' or '
+        if lg_event = '' then
+            lg_event := 'insert or update or delete';
+        end if;
+
+        -- create trigger
+        lg_args := lg_args || _extra_args;
+        sql := 'create trigger ' || quote_ident(trigger_name)
+            || ' ' || lg_pos || ' ' || lg_event
+            || ' on ' || londiste.quote_fqname(i_dest_table)
+            || ' for each row execute procedure '
+            || lg_func || '(' || array_to_string(lg_args, ', ') || ')';
+        execute sql;
+    end if;
+
+    -- create truncate trigger if it does not exists already
+    show server_version_num into pgversion;
+    if pgversion >= 80400 then
+        trunctrg_name  := '_londiste_' || i_queue_name || '_truncate';
+        perform 1 from pg_catalog.pg_trigger
+          where tgrelid = londiste.find_table_oid(i_dest_table)
+            and tgname = trunctrg_name;
+        if not found then
+            _extra_args := quote_literal(i_queue_name) || _extra_args;
+            sql := 'create trigger ' || quote_ident(trunctrg_name)
+                || ' after truncate on ' || londiste.quote_fqname(i_dest_table)
+                || ' for each statement execute procedure pgq.sqltriga('
+                || array_to_string(_extra_args, ', ') || ')';
+            execute sql;
+        end if;
+    end if;
+
+    select 200, 'OK'
+    into ret_code, ret_note;
+    return;
+end;
+$$ language plpgsql;
+
+
 create or replace function londiste.local_add_table(
     in i_queue_name     text,
     in i_table_name     text,
@@ -472,91 +752,53 @@ as $$
 --      200 - Ok
 --      301 - Warning, trigger exists that will fire before londiste one
 --      400 - No such set
+--      410 - Table already exists but with different table_attrs
 ------------------------------------------------------------------------
 declare
     col_types text;
     fq_table_name text;
     new_state text;
-    trunctrg_name text;
     pgversion int;
     logtrg_previous text;
-    lg_name text;
-    lg_func text;
-    lg_pos text;
-    lg_event text;
-    lg_args text;
-    _extra_args text;
+    trigger_name text;
     tbl record;
     i integer;
     j integer;
-    sql text;
     arg text;
     _node record;
     _tbloid oid;
     _combined_queue text;
     _combined_table text;
-    -- skip trigger
-    _skip_prefix text := 'zzz_';
-    _skip_trg_count integer;
-    _skip_trg_name text;
+    _table_attrs text := i_table_attrs;
     -- check local tables from all sources
     _queue_name text;
     _local boolean;
-    -- array with all tgflags values
-    _check_flags char[] := array['B','A','Q','L','I','U','D','S'];
-    -- given tgflags array
-    _tgflags char[];
-    -- ordinary argument array
-    _args text[];
     -- argument flags
     _expect_sync boolean := false;
     _merge_all boolean := false;
     _no_merge boolean := false;
-    _skip_truncate boolean := false;
-    _no_triggers boolean := false;
-    _skip boolean := false;
     _virtual_table boolean := false;
     _dest_table text;
-    _got_extra1 boolean := false;
     _table_name2 text;
     _desc text;
 begin
 
-    -------- i_trg_args ARGUMENTS PARSING
+    -------- i_trg_args ARGUMENTS PARSING (TODO: use different input param for passing extra options that have nothing to do with trigger)
 
     if array_lower(i_trg_args, 1) is not null then
         for i in array_lower(i_trg_args, 1) .. array_upper(i_trg_args, 1) loop
             arg := i_trg_args[i];
-            if arg like 'tgflags=%' then
-                -- special flag handling
-                arg := upper(substr(arg, 9));
-                for j in array_lower(_check_flags, 1) .. array_upper(_check_flags, 1) loop
-                    if position(_check_flags[j] in arg) > 0 then
-                        _tgflags := array_append(_tgflags, _check_flags[j]);
-                    end if;
-                end loop;
-            elsif arg = 'expect_sync' then
+            if arg = 'expect_sync' then
                 _expect_sync := true;
             elsif arg = 'skip_truncate' then
-                _skip_truncate := true;
-            elsif arg = 'no_triggers' then
-                _no_triggers := true;
+                _table_attrs := coalesce(_table_attrs || '&skip_truncate=1', 'skip_truncate=1');
             elsif arg = 'merge_all' then
                 _merge_all = true;
             elsif arg = 'no_merge' then
                 _no_merge = true;
-            elsif lower(arg) = 'skip' then
-                _skip := true;
             elsif arg = 'virtual_table' then
                 _virtual_table := true;
                 _expect_sync := true;   -- do not copy
-                _no_triggers := true;   -- do not create triggers
-            else
-                if arg like 'ev_extra1=%' then
-                    _got_extra1 := true;
-                end if;
-                -- ordinary arg
-                _args = array_append(_args, quote_literal(arg));
             end if;
         end loop;
     end if;
@@ -570,13 +812,6 @@ begin
     fq_table_name := londiste.make_fqname(i_table_name);
     _dest_table := londiste.make_fqname(coalesce(i_dest_table, i_table_name));
 
-    if _dest_table <> fq_table_name and not _got_extra1 then
-        -- if renamed table, enforce trigger to put
-        -- global table name into extra1
-        arg := 'ev_extra1=' || quote_literal(fq_table_name);
-        _args := array_append(_args, quote_literal(arg));
-    end if;
-
     if _dest_table = fq_table_name then
         _desc := fq_table_name;
     else
@@ -619,7 +854,7 @@ begin
         return;
     end if;
 
-    select merge_state, local into tbl
+    select merge_state, local, table_attrs into tbl
         from londiste.table_info
         where queue_name = i_queue_name and table_name = fq_table_name;
     if not found then
@@ -637,13 +872,17 @@ begin
         end if;
 
         -- reload info
-        select merge_state, local into tbl
+        select merge_state, local, table_attrs into tbl
             from londiste.table_info
             where queue_name = i_queue_name and table_name = fq_table_name;
     end if;
 
     if tbl.local then
-        select 200, 'Table already added: ' || _desc into ret_code, ret_note;
+        if tbl.table_attrs is distinct from _table_attrs then
+            select 410, 'Table ' || _desc || ' already added, but with different args: ' || coalesce(tbl.table_attrs, '') into ret_code, ret_note;
+        else
+            select 200, 'Table already added: ' || _desc into ret_code, ret_note;
+        end if;
         return;
     end if;
 
@@ -661,7 +900,7 @@ begin
     update londiste.table_info
         set local = true,
             merge_state = new_state,
-            table_attrs = coalesce(i_table_attrs, table_attrs),
+            table_attrs = coalesce(_table_attrs, table_attrs),
             dest_table = nullif(_dest_table, fq_table_name)
         where queue_name = i_queue_name and table_name = fq_table_name;
     if not found then
@@ -706,7 +945,7 @@ begin
             update londiste.table_info
                set local = true,
                    merge_state = new_state,
-                   table_attrs = coalesce(i_table_attrs, table_attrs)
+                   table_attrs = coalesce(_table_attrs, table_attrs)
                where queue_name = _queue_name and table_name = _table_name2;
             if not found then
                 raise exception 'lost table: % on queue %', _table_name2, _queue_name;
@@ -725,7 +964,7 @@ begin
             into _combined_queue, _combined_table;
         if found and _combined_table is null then
             select f.ret_code, f.ret_note
-                from londiste.local_add_table(_combined_queue, fq_table_name, i_trg_args, i_table_attrs, _dest_table) f
+                from londiste.local_add_table(_combined_queue, fq_table_name, i_trg_args, _table_attrs, _dest_table) f
                 into ret_code, ret_note;
             if ret_code >= 300 then
                 return;
@@ -733,148 +972,32 @@ begin
         end if;
     end if;
 
-    if _skip_truncate then
-        perform 1
-        from londiste.local_set_table_attrs(i_queue_name, fq_table_name,
-            coalesce(i_table_attrs || '&skip_truncate=1', 'skip_truncate=1'));
-    end if;
+    -- create trigger
+    select f.ret_code, f.ret_note, f.trigger_name
+        from londiste.create_trigger(i_queue_name, fq_table_name, i_trg_args, _dest_table, _node.node_type) f
+        into ret_code, ret_note, trigger_name;
 
-    -------- TRIGGER LOGIC
-
-    -- new trigger
-    _extra_args := '';
-    lg_name := '_londiste_' || i_queue_name;
-    lg_func := 'pgq.logutriga';
-    lg_event := '';
-    lg_args := quote_literal(i_queue_name);
-    lg_pos := 'after';
-
-    if array_lower(_args, 1) is not null then
-        lg_args := lg_args || ', ' || array_to_string(_args, ', ');
-    end if;
-
-    if 'B' = any(_tgflags) then
-        lg_pos := 'before';
-    end if;
-    if 'A' = any(_tgflags)  then
-        lg_pos := 'after';
-    end if;
-    if 'Q' = any(_tgflags) then
-        lg_func := 'pgq.sqltriga';
-    end if;
-    if 'L' = any(_tgflags) then
-        lg_func := 'pgq.logutriga';
-    end if;
-    if 'I' = any(_tgflags) then
-        lg_event := lg_event || ' or insert';
-    end if;
-    if 'U' = any(_tgflags) then
-        lg_event := lg_event || ' or update';
-    end if;
-    if 'D' = any(_tgflags) then
-        lg_event := lg_event || ' or delete';
-    end if;
-    if 'S' = any(_tgflags) then
-        _skip := true;
-    end if;
-
-    if _node.node_type = 'leaf' then
-        -- on weird leafs the trigger funcs may not exist
-        perform 1 from pg_proc p join pg_namespace n on (n.oid = p.pronamespace)
-            where n.nspname = 'pgq' and p.proname in ('logutriga', 'sqltriga');
-        if not found then
-            select 200, 'Table added with no triggers: ' || _desc into ret_code, ret_note;
-            return;
-        end if;
-        -- on regular leaf, install deny trigger
-        _extra_args := ', ' || quote_literal('deny');
-    end if;
-
-    -- if skip param given, rename previous skip triggers and prefix current
-    if _skip then
-        -- get count and name of existing skip triggers
-        select count(*), min(t.tgname)
-        into _skip_trg_count, _skip_trg_name
-        from pg_catalog.pg_trigger t
-        where t.tgrelid = londiste.find_table_oid(_dest_table)
-            and position(E'\\000skip\\000' in lower(tgargs::text)) > 0;
-        -- if no previous skip triggers, prefix name and add SKIP to args
-        if _skip_trg_count = 0 then
-            lg_name := _skip_prefix || lg_name;
-            lg_args := lg_args || ', ' || quote_literal('SKIP');
-        -- if one previous skip trigger, check it's prefix and
-        -- do not use SKIP on current trigger
-        elsif _skip_trg_count = 1 then
-            -- if not prefixed then rename
-            if position(_skip_prefix in _skip_trg_name) != 1 then
-                sql := 'alter trigger ' || _skip_trg_name
-                    || ' on ' || londiste.quote_fqname(_dest_table)
-                    || ' rename to ' || _skip_prefix || _skip_trg_name;
-                execute sql;
-            end if;
-        else
-            select 405, 'Multiple SKIP triggers in table: ' || _desc
+    if ret_code > 299 then
+        ret_note := 'Trigger creation failed for table ' || _desc || ': ' || ret_note;
+        return;
+    elsif ret_code = 201 then
+        select 200, 'Table added with no triggers: ' || _desc
             into ret_code, ret_note;
-            return;
-        end if;
-    end if;
-
-    -- create Ins/Upd/Del trigger if it does not exists already
-    perform 1 from pg_catalog.pg_trigger
-        where tgrelid = londiste.find_table_oid(_dest_table)
-            and tgname = lg_name;
-    if not found then
-
-        if _no_triggers then
-            select 200, 'Table added with no triggers: ' || _desc
-            into ret_code, ret_note;
-            return;
-        end if;
-
-        -- finalize event
-        lg_event := substr(lg_event, 4);
-        if lg_event = '' then
-            lg_event := 'insert or update or delete';
-        end if;
-
-        -- create trigger
-        sql := 'create trigger ' || quote_ident(lg_name)
-            || ' ' || lg_pos || ' ' || lg_event
-            || ' on ' || londiste.quote_fqname(_dest_table)
-            || ' for each row execute procedure '
-            || lg_func || '(' || lg_args || _extra_args || ')';
-        execute sql;
-    end if;
-
-    -- create truncate trigger if it does not exists already
-    show server_version_num into pgversion;
-    if pgversion >= 80400 then
-        trunctrg_name  := '_londiste_' || i_queue_name || '_truncate';
-        perform 1 from pg_catalog.pg_trigger
-          where tgrelid = londiste.find_table_oid(_dest_table)
-            and tgname = trunctrg_name;
-        if not found then
-            sql := 'create trigger ' || quote_ident(trunctrg_name)
-                || ' after truncate on ' || londiste.quote_fqname(_dest_table)
-                || ' for each statement execute procedure pgq.sqltriga(' || quote_literal(i_queue_name)
-                || _extra_args || ')';
-            execute sql;
-        end if;
+        return;
     end if;
 
     -- Check that no trigger exists on the target table that will get fired
-    -- before londiste one (this could have londiste replicate data
-    -- out-of-order
+    -- before londiste one (this could have londiste replicate data out-of-order)
     --
-    -- Don't report all the trigger names, 8.3 does not have array_accum
-    -- available
+    -- Don't report all the trigger names, 8.3 does not have array_accum available.
 
+    show server_version_num into pgversion;
     if pgversion >= 90000 then
         select tg.tgname into logtrg_previous
         from pg_class r join pg_trigger tg on (tg.tgrelid = r.oid)
         where r.oid = londiste.find_table_oid(_dest_table)
           and not tg.tgisinternal
-          and tg.tgname < lg_name::name
+          and tg.tgname < trigger_name::name
           -- per-row AFTER trigger
           and (tg.tgtype & 3) = 1   -- bits: 0:ROW, 1:BEFORE
           -- current londiste
@@ -888,7 +1011,7 @@ begin
         from pg_class r join pg_trigger tg on (tg.tgrelid = r.oid)
         where r.oid = londiste.find_table_oid(_dest_table)
           and not tg.tgisconstraint
-          and tg.tgname < lg_name::name
+          and tg.tgname < trigger_name::name
           -- per-row AFTER trigger
           and (tg.tgtype & 3) = 1   -- bits: 0:ROW, 1:BEFORE
           -- current londiste
@@ -971,6 +1094,84 @@ end;
 $$ language plpgsql strict;
 
 
+create or replace function londiste.local_change_handler(
+    in i_queue_name     text,
+    in i_table_name     text,
+    in i_trg_args       text[],
+    in i_table_attrs    text,
+    out ret_code        int4,
+    out ret_note        text)
+as $$
+----------------------------------------------------------------------------------------------------
+-- Function: londiste.local_change_handler(4)
+--
+--     Change handler and rebuild trigger if needed
+--
+-- Parameters:
+--      i_queue_name  - set name
+--      i_table_name  - table name
+--      i_trg_args    - args to trigger
+--      i_table_attrs - args to python handler
+--
+-- Returns:
+--      200 - OK
+--      400 - No such set
+--      404 - Table not found
+--
+----------------------------------------------------------------------------------------------------
+declare
+    _dest_table text;
+    _desc text;
+    _node record;
+begin
+    -- get node info
+    select * from pgq_node.get_node_info(i_queue_name) into _node;
+    if not found or _node.ret_code >= 400 then
+        select 400, 'No such set: ' || i_queue_name into ret_code, ret_note;
+        return;
+    end if;
+
+    -- update table_attrs with new handler info
+    select f.ret_code, f.ret_note
+        from londiste.local_set_table_attrs(i_queue_name, i_table_name, i_table_attrs) f
+        into ret_code, ret_note;
+    if ret_code <> 200 then
+        return;
+    end if;
+
+    -- get destination table name for use in trigger creation
+    select coalesce(ti.dest_table, i_table_name)
+        from londiste.table_info ti
+        where queue_name = i_queue_name
+        and table_name = i_table_name
+        and local
+        into _dest_table;
+
+    -- replace the trigger if needed
+    select f.ret_code, f.ret_note
+        from londiste.create_trigger(i_queue_name, i_table_name, i_trg_args, _dest_table, _node.node_type) f
+        into ret_code, ret_note;
+
+    if _dest_table = i_table_name then
+        _desc := i_table_name;
+    else
+        _desc := i_table_name || '(' || _dest_table || ')';
+    end if;
+
+    if ret_code > 299 then
+        ret_note := 'Trigger creation failed for table ' || _desc || ': ' || ret_note;
+        return;
+    elsif ret_code = 201 then
+        select 200, 'Table handler updated with no triggers: ' || _desc
+            into ret_code, ret_note;
+        return;
+    end if;
+
+    select 200, 'Handler changed for table: ' || _desc
+        into ret_code, ret_note;
+    return;
+end;
+$$ language plpgsql;
 
 
 
@@ -2156,7 +2357,7 @@ begin
     if TG_OP = 'DELETE' then
         perform londiste.drop_table_triggers(OLD.queue_name, OLD.table_name);
     end if;
-    return null;
+    return OLD;
 end;
 $$ language plpgsql;
 
@@ -2242,11 +2443,11 @@ returns text as $$
 -- ----------------------------------------------------------------------
 -- Function: londiste.version(0)
 --
---      Returns version string for londiste.  ATM it is based on SkyTools version
---      and only bumped when database code changes.
+--      Returns version string for londiste.  ATM it is based on SkyTools
+--      version and only bumped when database code changes.
 -- ----------------------------------------------------------------------
 begin
-    return '3.1.1';
+    return '3.2.4';
 end;
 $$ language plpgsql;
 
@@ -2264,10 +2465,10 @@ create or replace function londiste.create_partition(
     i_part_period text
 ) returns int as $$
 ------------------------------------------------------------------------
--- Function: public.create_partition
+-- Function: londiste.create_partition
 --
 --      Creates inherited child table if it does not exist by copying parent table's structure.
---      Locks parent table to avoid parallel creation.
+--      Locks londiste.table_info table to avoid parallel creation of any partitions.
 --
 -- Elements that are copied over by "LIKE x INCLUDING ALL":
 --      * Defaults
@@ -2303,6 +2504,7 @@ declare
     parent_oid      oid;
     part_schema     text;
     part_name       text;
+    owner           name;
     pos             int4;
     fq_table        text;
     fq_part         text;
@@ -2346,7 +2548,8 @@ begin
     fq_part := quote_ident(part_schema) || '.' || quote_ident(part_name);
 
     -- allow only single creation at a time, without affecting DML operations
-    execute 'lock table ' || fq_table || ' in share update exclusive mode';
+    -- (changed from locking parent table to avoid deadlocks from concurrent workers)
+    execute 'lock table londiste.table_info in share update exclusive mode';
     parent_oid := fq_table::regclass::oid;
 
     -- check if part table exists
@@ -2368,6 +2571,20 @@ begin
     sql := sql || ') inherits (' || fq_table || ')';
     execute sql;
 
+    -- find out parent table owner
+    select o.rolname into owner
+      from pg_class t, pg_namespace s, pg_roles o
+     where t.relnamespace = s.oid
+       and s.nspname = parent_schema
+       and t.relname = parent_name
+       and t.relowner = o.oid;
+
+    -- set proper part table ownership
+    if owner != user then
+        sql = 'alter table ' || fq_part || ' owner to ' || quote_ident(owner);
+        execute sql;
+    end if;
+
     -- extra check constraint
     if i_part_field != '' then
         part_start := date_trunc(i_part_period, i_part_time);
@@ -2439,7 +2656,7 @@ begin
 
     -- copy rules
     for r in
-        select rw.rulename, rw.ev_enabled, pg_get_ruledef(rw.oid) as definition
+        select rw.rulename, rw.ev_enabled, pg_catalog.pg_get_ruledef(rw.oid) as definition
           from pg_catalog.pg_rewrite rw
          where rw.ev_class = parent_oid
            and rw.rulename <> '_RETURN'::name
@@ -2476,6 +2693,7 @@ begin
         if r_extra is not null then
             sql := 'ALTER TABLE ' || fq_part || r_extra
                 || quote_ident(r.rulename);
+            execute sql;
         end if;
     end loop;
 
@@ -2485,6 +2703,162 @@ $$ language plpgsql;
 
 
 
+create or replace function londiste.is_obsolete_partition
+(
+    in i_partition_table text,
+    in i_retention_period interval,
+    in i_partition_period text
+)
+    returns boolean
+as $$
+-------------------------------------------------------------------------------
+--  Function: londiste.is_obsolete_partition(3)
+--
+--    Test partition name of partition-by-date parent table.
+--
+--  Parameters:
+--    i_partition_table     Partition table name we want to check
+--    i_retention_period    How long to keep partitions around
+--    i_partition_period    One of: year, month, day, hour
+--
+--  Returns:
+--    True if partition is too old, false if it is not,
+--    null if its name does not match expected pattern.
+-------------------------------------------------------------------------------
+declare
+    _expr text;
+    _dfmt text;
+    _base text;
+begin
+    if i_partition_period in ('year', 'yearly') then
+        _expr := '_[0-9]{4}';
+        _dfmt := '_YYYY';
+    elsif i_partition_period in ('month', 'monthly') then
+        _expr := '_[0-9]{4}_[0-9]{2}';
+        _dfmt := '_YYYY_MM';
+    elsif i_partition_period in ('day', 'daily') then
+        _expr := '_[0-9]{4}_[0-9]{2}_[0-9]{2}';
+        _dfmt := '_YYYY_MM_DD';
+    elsif i_partition_period in ('hour', 'hourly') then
+        _expr := '_[0-9]{4}_[0-9]{2}_[0-9]{2}_[0-9]{2}';
+        _dfmt := '_YYYY_MM_DD_HH24';
+    else
+        raise exception 'not supported i_partition_period: %', i_partition_period;
+    end if;
+
+    _expr = '^(.+)' || _expr || '$';
+    _base = substring (i_partition_table from _expr);
+
+    if _base is null then
+        return null;
+    elsif i_partition_table < _base || to_char (now() - i_retention_period, _dfmt) then
+        return true;
+    else
+        return false;
+    end if;
+end;
+$$ language plpgsql;
+
+
+
+create or replace function londiste.list_obsolete_partitions
+(
+    in i_parent_table text,
+    in i_retention_period interval,
+    in i_partition_period text
+)
+    returns setof text
+as $$
+-------------------------------------------------------------------------------
+--  Function: londiste.list_obsolete_partitions(3)
+--
+--    List obsolete partitions of partition-by-date parent table.
+--
+--  Parameters:
+--    i_parent_table        Master table from which partitions are inherited
+--    i_retention_period    How long to keep partitions around
+--    i_partition_period    One of: year, month, day, hour
+--
+--  Returns:
+--    Names of partitions to be dropped
+-------------------------------------------------------------------------------
+declare
+    _schema text not null := split_part (i_parent_table, '.', 1);
+    _table  text not null := split_part (i_parent_table, '.', 2);
+    _part   text;
+    _expr   text;
+    _dfmt   text;
+begin
+    if i_partition_period in ('year', 'yearly') then
+        _expr := '_[0-9]{4}';
+        _dfmt := '_YYYY';
+    elsif i_partition_period in ('month', 'monthly') then
+        _expr := '_[0-9]{4}_[0-9]{2}';
+        _dfmt := '_YYYY_MM';
+    elsif i_partition_period in ('day', 'daily') then
+        _expr := '_[0-9]{4}_[0-9]{2}_[0-9]{2}';
+        _dfmt := '_YYYY_MM_DD';
+    elsif i_partition_period in ('hour', 'hourly') then
+        _expr := '_[0-9]{4}_[0-9]{2}_[0-9]{2}_[0-9]{2}';
+        _dfmt := '_YYYY_MM_DD_HH24';
+    else
+        raise exception 'not supported i_partition_period: %', i_partition_period;
+    end if;
+
+    if length (_table) = 0 then
+        _table := _schema;
+        _schema := 'public';
+    end if;
+
+    for _part in
+        select quote_ident (t.schemaname) ||'.'|| quote_ident (t.tablename)
+          from pg_catalog.pg_tables t
+         where t.schemaname = _schema
+           and t.tablename ~ ('^'|| _table || _expr ||'$')
+           and t.tablename < _table || to_char (now() - i_retention_period, _dfmt)
+         order by 1
+    loop
+        return next _part;
+    end loop;
+end;
+$$ language plpgsql;
+
+
+
+create or replace function londiste.drop_obsolete_partitions
+(
+    in i_parent_table text,
+    in i_retention_period interval,
+    in i_partition_period text
+)
+    returns setof text
+as $$
+-------------------------------------------------------------------------------
+--  Function: londiste.drop_obsolete_partitions(3)
+--
+--    Drop obsolete partitions of partition-by-date parent table.
+--
+--  Parameters:
+--    i_parent_table        Master table from which partitions are inherited
+--    i_retention_period    How long to keep partitions around
+--    i_partition_period    One of: year, month, day, hour
+--
+--  Returns:
+--    Names of partitions dropped
+-------------------------------------------------------------------------------
+declare
+    _part text;
+begin
+    for _part in
+        select londiste.list_obsolete_partitions (i_parent_table, i_retention_period, i_partition_period)
+    loop
+        execute 'drop table '|| _part;
+        return next _part;
+    end loop;
+end;
+$$ language plpgsql;
+
+
 
 
 
