@@ -153,7 +153,7 @@ begin
     for rec in
         select xtbl from pgq.batch_event_tables(x_batch_id) xtbl
     loop
-        tbl := rec.xtbl;
+        tbl := pgq.quote_fqname(rec.xtbl);
         -- this gets newer queries that definitely are not in prev_snapshot
         part := select_fields
             || ' from pgq.tick cur, pgq.tick last, ' || tbl || ' ev '
@@ -729,8 +729,8 @@ begin
     -- there may be long lock on the table from pg_dump,
     -- detect it and skip rotate then
     begin
-        execute 'lock table ' || tbl || ' nowait';
-        execute 'truncate ' || tbl;
+        execute 'lock table ' || pgq.quote_fqname(tbl) || ' nowait';
+        execute 'truncate ' || pgq.quote_fqname(tbl);
     exception
         when lock_not_available then
             -- cannot truncate, skipping rotate
@@ -1013,11 +1013,12 @@ begin
     end if;
 
     -- tick seq, normal users don't need to modify it
-    execute 'grant select on ' || q.queue_tick_seq || ' to public';
+    execute 'grant select on ' || pgq.quote_fqname(q.queue_tick_seq) || ' to public';
 
     -- event seq
-    execute 'grant select on ' || q.queue_event_seq || ' to public';
-    
+    execute 'grant select on ' || pgq.quote_fqname(q.queue_event_seq) || ' to public';
+    execute 'grant usage on ' || pgq.quote_fqname(q.queue_event_seq) || ' to pgq_admin';
+
     -- set grants on parent table
     perform pgq._grant_perms_from('pgq', 'event_template', dst_schema, dst_table);
 
@@ -1069,7 +1070,7 @@ begin
 
     return 1;
 end;
-$$ language plpgsql;
+$$ language plpgsql strict;
 
 
 
@@ -1208,7 +1209,7 @@ begin
     execute 'select last_value from ' || fqname into res;
     return res;
 end;
-$$ language plpgsql;
+$$ language plpgsql strict;
 
 create or replace function pgq.seq_setval(i_seq_name text, i_new_value int8)
 returns bigint as $$
@@ -1227,19 +1228,8 @@ returns bigint as $$
 declare
     res     int8;
     fqname  text;
-    pos     integer;
-    s       text;
-    n       text;
 begin
-    pos := position('.' in i_seq_name);
-    if pos > 0 then
-        s := substring(i_seq_name for pos - 1);
-        n := substring(i_seq_name from pos + 1);
-    else
-        s := 'public';
-        n := i_seq_name;
-    end if;
-    fqname := quote_ident(s) || '.' || quote_ident(n);
+    fqname := pgq.quote_fqname(i_seq_name);
 
     res := pgq.seq_getval(i_seq_name);
     if res < i_new_value then
@@ -1248,7 +1238,45 @@ begin
     end if;
     return res;
 end;
-$$ language plpgsql;
+$$ language plpgsql strict;
+
+
+
+
+create or replace function pgq.quote_fqname(i_name text)
+returns text as $$
+-- ----------------------------------------------------------------------
+-- Function: pgq.quote_fqname(1)
+--
+--      Quete fully-qualified object name for SQL.
+--
+--      First dot is taken as schema separator.
+--
+--      If schema is missing, 'public' is assumed.
+--
+-- Parameters:
+--      i_name  - fully qualified object name.
+--
+-- Returns:
+--      Quoted name.
+-- ----------------------------------------------------------------------
+declare
+    res     text;
+    pos     integer;
+    s       text;
+    n       text;
+begin
+    pos := position('.' in i_name);
+    if pos > 0 then
+        s := substring(i_name for pos - 1);
+        n := substring(i_name from pos + 1);
+    else
+        s := 'public';
+        n := i_name;
+    end if;
+    return quote_ident(s) || '.' || quote_ident(n);
+end;
+$$ language plpgsql strict immutable;
 
 
 
@@ -1349,21 +1377,21 @@ begin
         where queue_id = id;
 
     -- create seqs
-    execute 'CREATE SEQUENCE ' || tick_seq;
-    execute 'CREATE SEQUENCE ' || ev_seq;
+    execute 'CREATE SEQUENCE ' || pgq.quote_fqname(tick_seq);
+    execute 'CREATE SEQUENCE ' || pgq.quote_fqname(ev_seq);
 
     -- create data tables
-    execute 'CREATE TABLE ' || tblpfx || ' () '
+    execute 'CREATE TABLE ' || pgq.quote_fqname(tblpfx) || ' () '
             || ' INHERITS (pgq.event_template)';
     for i in 0 .. (n_tables - 1) loop
         tblname := tblpfx || '_' || i::text;
-        idxname := idxpfx || '_' || i::text;
-        execute 'CREATE TABLE ' || tblname || ' () '
-                || ' INHERITS (' || tblpfx || ')';
-        execute 'ALTER TABLE ' || tblname || ' ALTER COLUMN ev_id '
+        idxname := idxpfx || '_' || i::text || '_txid_idx';
+        execute 'CREATE TABLE ' || pgq.quote_fqname(tblname) || ' () '
+                || ' INHERITS (' || pgq.quote_fqname(tblpfx) || ')';
+        execute 'ALTER TABLE ' || pgq.quote_fqname(tblname) || ' ALTER COLUMN ev_id '
                 || ' SET DEFAULT nextval(' || quote_literal(ev_seq) || ')';
-        execute 'create index ' || idxname || '_txid_idx on '
-                || tblname || ' (ev_txid)';
+        execute 'create index ' || quote_ident(idxname) || ' on '
+                || pgq.quote_fqname(tblname) || ' (ev_txid)';
     end loop;
 
     perform pgq.grant_perms(i_queue_name);
@@ -1426,17 +1454,17 @@ begin
     -- drop data tables
     for i in 0 .. (q.queue_ntables - 1) loop
         tblname := q.queue_data_pfx || '_' || i::text;
-        execute 'DROP TABLE ' || tblname;
+        execute 'DROP TABLE ' || pgq.quote_fqname(tblname);
     end loop;
-    execute 'DROP TABLE ' || q.queue_data_pfx;
+    execute 'DROP TABLE ' || pgq.quote_fqname(q.queue_data_pfx);
 
     -- delete ticks
     delete from pgq.tick where tick_queue = q.queue_id;
 
     -- drop seqs
     -- FIXME: any checks needed here?
-    execute 'DROP SEQUENCE ' || q.queue_tick_seq;
-    execute 'DROP SEQUENCE ' || q.queue_event_seq;
+    execute 'DROP SEQUENCE ' || pgq.quote_fqname(q.queue_tick_seq);
+    execute 'DROP SEQUENCE ' || pgq.quote_fqname(q.queue_event_seq);
 
     -- delete event
     delete from pgq.queue
@@ -1689,7 +1717,7 @@ returns integer as $$
 declare
     tmp         text;
     last_tick   bigint;
-    x_queue_id          integer;
+    x_queue_id  integer;
     x_consumer_id integer;
     queue integer;
     sub record;
@@ -1702,7 +1730,8 @@ begin
 
     -- get consumer and create if new
     select co_id into x_consumer_id from pgq.consumer
-        where co_name = x_consumer_name;
+        where co_name = x_consumer_name
+        for update;
     if not found then
         insert into pgq.consumer (co_name) values (x_consumer_name);
         x_consumer_id := currval('pgq.consumer_co_id_seq');
@@ -1770,8 +1799,8 @@ returns integer as $$
 -- ----------------------------------------------------------------------
 -- Function: pgq.unregister_consumer(2)
 --
---      Unsubscriber consumer from the queue.  Also consumer's
---      retry events are deleted.
+--      Unsubscribe consumer from the queue.
+--      Also consumer's retry events are deleted.
 --
 -- Parameters:
 --      x_queue_name        - Name of the queue
@@ -1788,14 +1817,19 @@ returns integer as $$
 declare
     x_sub_id integer;
     _sub_id_cnt integer;
+    _consumer_id integer;
+    _is_subconsumer boolean;
 begin
-    select s.sub_id into x_sub_id
-        from pgq.subscription s, pgq.consumer c, pgq.queue q
-        where s.sub_queue = q.queue_id
-          and s.sub_consumer = c.co_id
-          and q.queue_name = x_queue_name
-          and c.co_name = x_consumer_name
-        for update of s;
+    select s.sub_id, c.co_id,
+           -- subconsumers can only have both null or both not null - main consumer for subconsumers has only one not null
+           (s.sub_last_tick IS NULL AND s.sub_next_tick IS NULL) OR (s.sub_last_tick IS NOT NULL AND s.sub_next_tick IS NOT NULL)
+      into x_sub_id, _consumer_id, _is_subconsumer
+      from pgq.subscription s, pgq.consumer c, pgq.queue q
+     where s.sub_queue = q.queue_id
+       and s.sub_consumer = c.co_id
+       and q.queue_name = x_queue_name
+       and c.co_name = x_consumer_name
+       for update of s, c;
     if not found then
         return 0;
     end if;
@@ -1805,15 +1839,33 @@ begin
         from pgq.subscription
        where sub_id = x_sub_id;
 
-    -- retry events
-    delete from pgq.retry_queue
-        where ev_owner = x_sub_id;
+    -- delete only one subconsumer
+    if _sub_id_cnt > 1 and _is_subconsumer then
+        delete from pgq.subscription
+              where sub_id = x_sub_id
+                and sub_consumer = _consumer_id;
+        return 1;
+    else
+        -- delete main consumer (including possible subconsumers)
 
-    -- this will drop subconsumers too
-    delete from pgq.subscription
-        where sub_id = x_sub_id;
+        -- retry events
+        delete from pgq.retry_queue
+            where ev_owner = x_sub_id;
+
+        -- this will drop subconsumers too
+        delete from pgq.subscription
+            where sub_id = x_sub_id;
+
+        perform 1 from pgq.subscription
+            where sub_consumer = _consumer_id;
+        if not found then
+            delete from pgq.consumer
+                where co_id = _consumer_id;
+        end if;
+
+        return _sub_id_cnt;
+    end if;
 
-    return _sub_id_cnt;
 end;
 $$ language plpgsql security definer;
 
@@ -2664,11 +2716,11 @@ returns text as $$
 -- ----------------------------------------------------------------------
 -- Function: pgq.version(0)
 --
---      Returns version string for pgq.  ATM it is based on SkyTools version
---      and only bumped when database code changes.
+--      Returns version string for pgq.  ATM it is based on SkyTools
+--      version and only bumped when database code changes.
 -- ----------------------------------------------------------------------
 begin
-    return '3.1.1';
+    return '3.2.6';
 end;
 $$ language plpgsql;
 
@@ -2874,7 +2926,7 @@ SELECT pg_catalog.pg_extension_config_dump('pgq.subscription', '');
 SELECT pg_catalog.pg_extension_config_dump('pgq.event_template', '');
 SELECT pg_catalog.pg_extension_config_dump('pgq.retry_queue', '');
 
----- pg_dump is broken and cannot handle dumpable sequences
--- SELECT pg_catalog.pg_extension_config_dump('pgq.batch_id_seq', '');
+-- This needs pg_dump 9.1.7+
+SELECT pg_catalog.pg_extension_config_dump('pgq.batch_id_seq', '');
 
 
