@@ -593,7 +593,7 @@ returns text as $$
 --      only bumped when database code changes.
 -- ----------------------------------------------------------------------
 begin
-    return '3.1.0.0';
+    return '3.1';
 end;
 $$ language plpgsql;
 
