@@ -1,135 +1,5 @@
 
 
-
-
--- ----------------------------------------------------------------------
--- File: Tables
---
---      Schema 'pgq_node', contains tables for cascaded pgq.
---
--- Event types for cascaded queue:
---      pgq.location-info       - ev_data: node_name, extra1: queue_name, extra2: location, extra3: dead
---                                It contains updated node connect string.
---
---      pgq.global-watermark    - ev_data: tick_id,  extra1: queue_name
---                                Root node sends minimal tick_id that must be kept.
---
---      pgq.tick-id             - ev_data: tick_id,  extra1: queue_name
---                                Partition node inserts it's tick-id into combined queue.
---
--- ----------------------------------------------------------------------
-
-create schema pgq_node;
-
--- ----------------------------------------------------------------------
--- Table: pgq_node.location
---
---      Static table that just lists all members in set.
---
--- Columns:
---      queue_name      - cascaded queue name
---      node_name       - node name
---      node_location   - libpq connect string for connecting to node
---      dead            - whether the node is offline
--- ----------------------------------------------------------------------
-create table pgq_node.node_location (
-    queue_name      text not null,
-    node_name       text not null,
-    node_location   text not null,
-    dead            boolean not null default false,
-
-    primary key (queue_name, node_name)
-);
-
--- ----------------------------------------------------------------------
--- Table: pgq_node.node_info
---
---      Local node info.
---
--- Columns:
---      queue_name          - cascaded queue name
---      node_type           - local node type
---      node_name           - local node name
---      provider_node       - provider node name
---      worker_name         - consumer name that maintains this node
---      combined_queue      - on 'leaf' the target combined set name
---      node_attrs          - urlencoded fields for worker
---
--- Node types:
---      root            - data + batches is generated here
---      branch          - replicates full queue contents and maybe contains some tables
---      leaf            - does not replicate queue / or uses combined queue for that
--- ----------------------------------------------------------------------
-create table pgq_node.node_info (
-    queue_name      text not null primary key,
-    node_type       text not null,
-    node_name       text not null,
-    worker_name     text,
-    combined_queue  text,
-    node_attrs      text,
-
-    foreign key (queue_name, node_name) references pgq_node.node_location,
-    check (node_type in ('root', 'branch', 'leaf')),
-    check (case when node_type = 'root'   then  (worker_name is not null and combined_queue is null)
-                when node_type = 'branch' then  (worker_name is not null and combined_queue is null)
-                when node_type = 'leaf'   then  (worker_name is not null)
-                else false end)
-);
-
--- ----------------------------------------------------------------------
--- Table: pgq_node.local_state
---
---      All cascaded consumers (both worker and non-worker)
---      keep their state here.
---
--- Columns:
---      queue_name      - cascaded queue name
---      consumer_name   - cascaded consumer name
---      provider_node   - node name the consumer reads from
---      last_tick_id    - last committed tick id on this node
---      cur_error       - reason why current batch failed
---      paused          - whether consumer should wait
---      uptodate        - if consumer has seen new state
--- ----------------------------------------------------------------------
-create table pgq_node.local_state (
-    queue_name      text not null,
-    consumer_name   text not null,
-    provider_node   text not null,
-    last_tick_id    bigint not null,
-    cur_error       text,
-
-    paused          boolean not null default false,
-    uptodate        boolean not null default false,
-
-    primary key (queue_name, consumer_name),
-    foreign key (queue_name) references pgq_node.node_info,
-    foreign key (queue_name, provider_node) references pgq_node.node_location
-);
-
--- ----------------------------------------------------------------------
--- Table: pgq_node.subscriber
---
---      List of nodes that subscribe to local node.
---
--- Columns:
---      queue_name      - cascaded queue name
---      subscriber_node - node name that uses this node as provider.
---      worker_name     - consumer name that maintains remote node
--- ----------------------------------------------------------------------
-create table pgq_node.subscriber_info (
-    queue_name          text not null,
-    subscriber_node     text not null,
-    worker_name         text not null,
-    watermark_name      text not null,
-
-    primary key (queue_name, subscriber_node),
-    foreign key (queue_name, subscriber_node) references pgq_node.node_location,
-    foreign key (worker_name) references pgq.consumer (co_name),
-    foreign key (watermark_name) references pgq.consumer (co_name)
-);
-
-
-
 -- File: Functions
 --
 --      Database functions for cascaded pgq.
@@ -626,11 +496,12 @@ create or replace function pgq_node.get_node_info(
 --      provider_node - provider node name
 --      provider_location - provider connect string
 --      combined_queue - queue name for target set
---      combined_type - node type of target setA
+--      combined_type - node type of target set
 --      worker_name - consumer name that maintains this node
 --      worker_paused - is worker paused
 --      worker_uptodate - is worker seen the changes
 --      worker_last_tick - last committed tick_id by worker
+--      node_attrs - urlencoded dict of random attrs for worker (eg. sync_watermark)
 -- ----------------------------------------------------------------------
 declare
     sql text;
@@ -866,7 +737,8 @@ declare
 begin
     select node_type, worker_name into n_type, w_name
         from pgq_node.node_info
-        where queue_name = i_queue_name;
+        where queue_name = i_queue_name
+        for update;
     if not found then
         select 404, 'Node not initialized for queue: ' || i_queue_name
           into ret_code, ret_note;
@@ -975,7 +847,8 @@ declare
 begin
     select node_name, node_type, worker_name into n_name, n_type, w_name
         from pgq_node.node_info
-        where queue_name = i_queue_name;
+        where queue_name = i_queue_name
+        for update;
     if not found then
         select 404, 'Node not initialized for queue: ' || i_queue_name
           into ret_code, ret_note;
@@ -1161,13 +1034,13 @@ begin
         return;
     end if;
 
-    perform pgq.unregister_consumer(i_queue_name, n_wm_name);
-    perform pgq.unregister_consumer(i_queue_name, worker_name);
-
     delete from pgq_node.subscriber_info
         where queue_name = i_queue_name
             and subscriber_node = i_remote_node_name;
 
+    perform pgq.unregister_consumer(i_queue_name, n_wm_name);
+    perform pgq.unregister_consumer(i_queue_name, worker_name);
+
     select 200, 'Subscriber unregistered: '||i_remote_node_name
         into ret_code, ret_note;
     return;
@@ -1962,11 +1835,11 @@ returns text as $$
 -- ----------------------------------------------------------------------
 -- Function: pgq_node.version(0)
 --
---      Returns version string for pgq_node.  ATM it is based on SkyTools version
---      and only bumped when database code changes.
+--      Returns version string for pgq_node.  ATM it is based on SkyTools
+--      version and only bumped when database code changes.
 -- ----------------------------------------------------------------------
 begin
-    return '3.1.0.0';
+    return '3.2.5';
 end;
 $$ language plpgsql;
 
@@ -1974,21 +1847,3 @@ $$ language plpgsql;
 
 
 
-
-grant usage on schema pgq_node to public;
-
-
-
-
-
-
-
--- tag data objects as dumpable
-
-SELECT pg_catalog.pg_extension_config_dump('pgq_node.node_location', '');
-SELECT pg_catalog.pg_extension_config_dump('pgq_node.node_info', '');
-SELECT pg_catalog.pg_extension_config_dump('pgq_node.local_state', '');
-SELECT pg_catalog.pg_extension_config_dump('pgq_node.subscriber_info', '');
-
-
-
