@@ -497,11 +497,12 @@ create or replace function pgq_node.get_node_info(
 --      provider_node - provider node name
 --      provider_location - provider connect string
 --      combined_queue - queue name for target set
---      combined_type - node type of target setA
+--      combined_type - node type of target set
 --      worker_name - consumer name that maintains this node
 --      worker_paused - is worker paused
 --      worker_uptodate - is worker seen the changes
 --      worker_last_tick - last committed tick_id by worker
+--      node_attrs - urlencoded dict of random attrs for worker (eg. sync_watermark)
 -- ----------------------------------------------------------------------
 declare
     sql text;
@@ -737,7 +738,8 @@ declare
 begin
     select node_type, worker_name into n_type, w_name
         from pgq_node.node_info
-        where queue_name = i_queue_name;
+        where queue_name = i_queue_name
+        for update;
     if not found then
         select 404, 'Node not initialized for queue: ' || i_queue_name
           into ret_code, ret_note;
@@ -846,7 +848,8 @@ declare
 begin
     select node_name, node_type, worker_name into n_name, n_type, w_name
         from pgq_node.node_info
-        where queue_name = i_queue_name;
+        where queue_name = i_queue_name
+        for update;
     if not found then
         select 404, 'Node not initialized for queue: ' || i_queue_name
           into ret_code, ret_note;
@@ -1032,13 +1035,13 @@ begin
         return;
     end if;
 
-    perform pgq.unregister_consumer(i_queue_name, n_wm_name);
-    perform pgq.unregister_consumer(i_queue_name, worker_name);
-
     delete from pgq_node.subscriber_info
         where queue_name = i_queue_name
             and subscriber_node = i_remote_node_name;
 
+    perform pgq.unregister_consumer(i_queue_name, n_wm_name);
+    perform pgq.unregister_consumer(i_queue_name, worker_name);
+
     select 200, 'Subscriber unregistered: '||i_remote_node_name
         into ret_code, ret_note;
     return;
@@ -1833,11 +1836,11 @@ returns text as $$
 -- ----------------------------------------------------------------------
 -- Function: pgq_node.version(0)
 --
---      Returns version string for pgq_node.  ATM it is based on SkyTools version
---      and only bumped when database code changes.
+--      Returns version string for pgq_node.  ATM it is based on SkyTools
+--      version and only bumped when database code changes.
 -- ----------------------------------------------------------------------
 begin
-    return '3.1.0.0';
+    return '3.2.5';
 end;
 $$ language plpgsql;
 
